import { produce } from 'immer';
import upperFirst from 'lodash.upperfirst';
import { createSelector } from 'reselect';
import { asClass, Lifetime, createContainer, InjectionMode } from 'awilix';

function _extends() {
  _extends = Object.assign ? Object.assign.bind() : function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];
      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }
    return target;
  };
  return _extends.apply(this, arguments);
}
function _inheritsLoose(subClass, superClass) {
  subClass.prototype = Object.create(superClass.prototype);
  subClass.prototype.constructor = subClass;
  _setPrototypeOf(subClass, superClass);
}
function _setPrototypeOf(o, p) {
  _setPrototypeOf = Object.setPrototypeOf ? Object.setPrototypeOf.bind() : function _setPrototypeOf(o, p) {
    o.__proto__ = p;
    return o;
  };
  return _setPrototypeOf(o, p);
}
function _assertThisInitialized(self) {
  if (self === void 0) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }
  return self;
}
var id = 0;
function _classPrivateFieldLooseKey(name) {
  return "__private_" + id++ + "_" + name;
}
function _classPrivateFieldLooseBase(receiver, privateKey) {
  if (!Object.prototype.hasOwnProperty.call(receiver, privateKey)) {
    throw new TypeError("attempted to use private field on non-instance");
  }
  return receiver;
}

/**
 * Adapted from React: https://github.com/facebook/react/blob/master/packages/shared/formatProdErrorMessage.js
 *
 * Do not require this module directly! Use normal throw error calls. These messages will be replaced with error codes
 * during build.
 * @param {number} code
 */
function formatProdErrorMessage(code) {
  return "Minified Redux error #" + code + "; visit https://redux.js.org/Errors?code=" + code + " for the full message or " + 'use the non-minified dev environment for full errors. ';
}

/**
 * These are private action types reserved by Redux.
 * For any unknown actions, you must return the current state.
 * If the current state is undefined, you must return the initial state.
 * Do not reference these action types directly in your code.
 */
var randomString = function randomString() {
  return Math.random().toString(36).substring(7).split('').join('.');
};

var ActionTypes = {
  INIT: "@@redux/INIT" + randomString(),
  REPLACE: "@@redux/REPLACE" + randomString(),
  PROBE_UNKNOWN_ACTION: function PROBE_UNKNOWN_ACTION() {
    return "@@redux/PROBE_UNKNOWN_ACTION" + randomString();
  }
};

/**
 * @param {any} obj The object to inspect.
 * @returns {boolean} True if the argument appears to be a plain object.
 */
function isPlainObject(obj) {
  if (typeof obj !== 'object' || obj === null) return false;
  var proto = obj;

  while (Object.getPrototypeOf(proto) !== null) {
    proto = Object.getPrototypeOf(proto);
  }

  return Object.getPrototypeOf(obj) === proto;
}

// Inlined / shortened version of `kindOf` from https://github.com/jonschlinkert/kind-of
function miniKindOf(val) {
  if (val === void 0) return 'undefined';
  if (val === null) return 'null';
  var type = typeof val;

  switch (type) {
    case 'boolean':
    case 'string':
    case 'number':
    case 'symbol':
    case 'function':
      {
        return type;
      }
  }

  if (Array.isArray(val)) return 'array';
  if (isDate(val)) return 'date';
  if (isError(val)) return 'error';
  var constructorName = ctorName(val);

  switch (constructorName) {
    case 'Symbol':
    case 'Promise':
    case 'WeakMap':
    case 'WeakSet':
    case 'Map':
    case 'Set':
      return constructorName;
  } // other


  return type.slice(8, -1).toLowerCase().replace(/\s/g, '');
}

function ctorName(val) {
  return typeof val.constructor === 'function' ? val.constructor.name : null;
}

function isError(val) {
  return val instanceof Error || typeof val.message === 'string' && val.constructor && typeof val.constructor.stackTraceLimit === 'number';
}

function isDate(val) {
  if (val instanceof Date) return true;
  return typeof val.toDateString === 'function' && typeof val.getDate === 'function' && typeof val.setDate === 'function';
}

function kindOf(val) {
  var typeOfVal = typeof val;

  if (process.env.NODE_ENV !== 'production') {
    typeOfVal = miniKindOf(val);
  }

  return typeOfVal;
}

/**
 * Prints a warning in the console if it exists.
 *
 * @param {String} message The warning message.
 * @returns {void}
 */
function warning(message) {
  /* eslint-disable no-console */
  if (typeof console !== 'undefined' && typeof console.error === 'function') {
    console.error(message);
  }
  /* eslint-enable no-console */


  try {
    // This error was thrown as a convenience so that if you enable
    // "break on all exceptions" in your console,
    // it would pause the execution at this line.
    throw new Error(message);
  } catch (e) {} // eslint-disable-line no-empty

}

function getUnexpectedStateShapeWarningMessage(inputState, reducers, action, unexpectedKeyCache) {
  var reducerKeys = Object.keys(reducers);
  var argumentName = action && action.type === ActionTypes.INIT ? 'preloadedState argument passed to createStore' : 'previous state received by the reducer';

  if (reducerKeys.length === 0) {
    return 'Store does not have a valid reducer. Make sure the argument passed ' + 'to combineReducers is an object whose values are reducers.';
  }

  if (!isPlainObject(inputState)) {
    return "The " + argumentName + " has unexpected type of \"" + kindOf(inputState) + "\". Expected argument to be an object with the following " + ("keys: \"" + reducerKeys.join('", "') + "\"");
  }

  var unexpectedKeys = Object.keys(inputState).filter(function (key) {
    return !reducers.hasOwnProperty(key) && !unexpectedKeyCache[key];
  });
  unexpectedKeys.forEach(function (key) {
    unexpectedKeyCache[key] = true;
  });
  if (action && action.type === ActionTypes.REPLACE) return;

  if (unexpectedKeys.length > 0) {
    return "Unexpected " + (unexpectedKeys.length > 1 ? 'keys' : 'key') + " " + ("\"" + unexpectedKeys.join('", "') + "\" found in " + argumentName + ". ") + "Expected to find one of the known reducer keys instead: " + ("\"" + reducerKeys.join('", "') + "\". Unexpected keys will be ignored.");
  }
}

function assertReducerShape(reducers) {
  Object.keys(reducers).forEach(function (key) {
    var reducer = reducers[key];
    var initialState = reducer(undefined, {
      type: ActionTypes.INIT
    });

    if (typeof initialState === 'undefined') {
      throw new Error(process.env.NODE_ENV === "production" ? formatProdErrorMessage(12) : "The slice reducer for key \"" + key + "\" returned undefined during initialization. " + "If the state passed to the reducer is undefined, you must " + "explicitly return the initial state. The initial state may " + "not be undefined. If you don't want to set a value for this reducer, " + "you can use null instead of undefined.");
    }

    if (typeof reducer(undefined, {
      type: ActionTypes.PROBE_UNKNOWN_ACTION()
    }) === 'undefined') {
      throw new Error(process.env.NODE_ENV === "production" ? formatProdErrorMessage(13) : "The slice reducer for key \"" + key + "\" returned undefined when probed with a random type. " + ("Don't try to handle '" + ActionTypes.INIT + "' or other actions in \"redux/*\" ") + "namespace. They are considered private. Instead, you must return the " + "current state for any unknown actions, unless it is undefined, " + "in which case you must return the initial state, regardless of the " + "action type. The initial state may not be undefined, but can be null.");
    }
  });
}
/**
 * Turns an object whose values are different reducer functions, into a single
 * reducer function. It will call every child reducer, and gather their results
 * into a single state object, whose keys correspond to the keys of the passed
 * reducer functions.
 *
 * @param {Object} reducers An object whose values correspond to different
 * reducer functions that need to be combined into one. One handy way to obtain
 * it is to use ES6 `import * as reducers` syntax. The reducers may never return
 * undefined for any action. Instead, they should return their initial state
 * if the state passed to them was undefined, and the current state for any
 * unrecognized action.
 *
 * @returns {Function} A reducer function that invokes every reducer inside the
 * passed object, and builds a state object with the same shape.
 */


function combineReducers(reducers) {
  var reducerKeys = Object.keys(reducers);
  var finalReducers = {};

  for (var i = 0; i < reducerKeys.length; i++) {
    var key = reducerKeys[i];

    if (process.env.NODE_ENV !== 'production') {
      if (typeof reducers[key] === 'undefined') {
        warning("No reducer provided for key \"" + key + "\"");
      }
    }

    if (typeof reducers[key] === 'function') {
      finalReducers[key] = reducers[key];
    }
  }

  var finalReducerKeys = Object.keys(finalReducers); // This is used to make sure we don't warn about the same
  // keys multiple times.

  var unexpectedKeyCache;

  if (process.env.NODE_ENV !== 'production') {
    unexpectedKeyCache = {};
  }

  var shapeAssertionError;

  try {
    assertReducerShape(finalReducers);
  } catch (e) {
    shapeAssertionError = e;
  }

  return function combination(state, action) {
    if (state === void 0) {
      state = {};
    }

    if (shapeAssertionError) {
      throw shapeAssertionError;
    }

    if (process.env.NODE_ENV !== 'production') {
      var warningMessage = getUnexpectedStateShapeWarningMessage(state, finalReducers, action, unexpectedKeyCache);

      if (warningMessage) {
        warning(warningMessage);
      }
    }

    var hasChanged = false;
    var nextState = {};

    for (var _i = 0; _i < finalReducerKeys.length; _i++) {
      var _key = finalReducerKeys[_i];
      var reducer = finalReducers[_key];
      var previousStateForKey = state[_key];
      var nextStateForKey = reducer(previousStateForKey, action);

      if (typeof nextStateForKey === 'undefined') {
        var actionType = action && action.type;
        throw new Error(process.env.NODE_ENV === "production" ? formatProdErrorMessage(14) : "When called with an action of type " + (actionType ? "\"" + String(actionType) + "\"" : '(unknown type)') + ", the slice reducer for key \"" + _key + "\" returned undefined. " + "To ignore an action, you must explicitly return the previous state. " + "If you want this reducer to hold no value, you can return null instead of undefined.");
      }

      nextState[_key] = nextStateForKey;
      hasChanged = hasChanged || nextStateForKey !== previousStateForKey;
    }

    hasChanged = hasChanged || finalReducerKeys.length !== Object.keys(state).length;
    return hasChanged ? nextState : state;
  };
}

var _store = /*#__PURE__*/_classPrivateFieldLooseKey("store");
var WithStore = /*#__PURE__*/function () {
  function WithStore() {
    Object.defineProperty(this, _store, {
      writable: true,
      value: null
    });
  }
  var _proto = WithStore.prototype;
  _proto.setStore = function setStore(store) {
    _classPrivateFieldLooseBase(this, _store)[_store] = store;
  };
  _proto.getStore = function getStore() {
    return _classPrivateFieldLooseBase(this, _store)[_store];
  };
  _proto.getState = function getState() {
    if (_classPrivateFieldLooseBase(this, _store)[_store]) {
      return _classPrivateFieldLooseBase(this, _store)[_store].getState();
    } else {
      throw new Error('Store did not inject.');
    }
  };
  _proto.dispatch = function dispatch(action) {
    if (_classPrivateFieldLooseBase(this, _store)[_store]) {
      return _classPrivateFieldLooseBase(this, _store)[_store].dispatch(action);
    } else {
      throw new Error('Store did not inject.');
    }
  };
  return WithStore;
}();

var savedReducers = null;
var ACTION_END = 'Action';
var DISPATCH_ACTION_KEY = '#__dispatchAction';
var RESET_STATE_ACTION_KEY = '#__resetAction';
var TRACE_KEY = 'Trace';
var _cases = /*#__PURE__*/_classPrivateFieldLooseKey("cases");
var _serviceName = /*#__PURE__*/_classPrivateFieldLooseKey("serviceName");
var BaseReducer = /*#__PURE__*/function (_WithStore) {
  _inheritsLoose(BaseReducer, _WithStore);
  function BaseReducer() {
    var _this;
    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }
    _this = _WithStore.call.apply(_WithStore, [this].concat(args)) || this;
    Object.defineProperty(_assertThisInitialized(_this), _cases, {
      writable: true,
      value: new Map()
    });
    Object.defineProperty(_assertThisInitialized(_this), _serviceName, {
      writable: true,
      value: null
    });
    return _this;
  }
  var _proto = BaseReducer.prototype;
  _proto.__registerAction = function __registerAction(action) {
    var _this2 = this;
    var type = _classPrivateFieldLooseBase(this, _serviceName)[_serviceName] + "::" + action.key;
    if (!_classPrivateFieldLooseBase(this, _cases)[_cases].get(type)) {
      _classPrivateFieldLooseBase(this, _cases)[_cases].set(type, function () {
        var _action$func;
        for (var _len2 = arguments.length, args = new Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
          args[_key2] = arguments[_key2];
        }
        return (_action$func = action.func).apply.apply(_action$func, [_this2].concat(args));
      });
    }
    var instance = Object.getPrototypeOf(this);
    instance[action.key] = function () {
      for (var _len3 = arguments.length, params = new Array(_len3), _key3 = 0; _key3 < _len3; _key3++) {
        params[_key3] = arguments[_key3];
      }
      _this2.dispatch({
        type: type,
        params: params
      });
    };
  };
  _proto.__registerActions = function __registerActions() {
    var _this3 = this;
    var instance = Object.getPrototypeOf(this);
    instance[DISPATCH_ACTION_KEY] = function (newStateFunc) {
      return newStateFunc;
    };
    instance[RESET_STATE_ACTION_KEY] = function () {
      return function () {
        return instance.getInitialState();
      };
    };
    var actionsNames = Object.getOwnPropertyNames(instance).filter(function (name) {
      return name.endsWith(ACTION_END);
    });
    actionsNames.forEach(function (name) {
      _this3.__registerAction({
        key: name,
        func: instance[name]
      });
    });
  };
  _proto.replaceReducer = function replaceReducer(serviceKey, initialReducers) {
    var _this4 = this,
      _extends2;
    _classPrivateFieldLooseBase(this, _serviceName)[_serviceName] = upperFirst(serviceKey);
    this.__registerActions();
    var instance = Object.getPrototypeOf(this);
    var reducer = function reducer() {
      var state = (arguments.length <= 0 ? undefined : arguments[0]) || instance.getInitialState();
      var action = arguments.length <= 1 ? undefined : arguments[1];
      var type = action.type,
        actionParams = action.params;
      var func;
      if (type.startsWith(_classPrivateFieldLooseBase(_this4, _serviceName)[_serviceName] + "::" + TRACE_KEY + "::")) {
        func = _classPrivateFieldLooseBase(_this4, _cases)[_cases].get(_classPrivateFieldLooseBase(_this4, _serviceName)[_serviceName] + "::" + DISPATCH_ACTION_KEY);
      } else if (type === _classPrivateFieldLooseBase(_this4, _serviceName)[_serviceName] + "::resetState") {
        func = _classPrivateFieldLooseBase(_this4, _cases)[_cases].get(_classPrivateFieldLooseBase(_this4, _serviceName)[_serviceName] + "::" + RESET_STATE_ACTION_KEY);
      } else {
        func = _classPrivateFieldLooseBase(_this4, _cases)[_cases].get(type);
      }
      return func ? produce(state, func.call(_this4, actionParams)) : state;
    };
    var allReducers = _extends({}, savedReducers || initialReducers, (_extends2 = {}, _extends2[_classPrivateFieldLooseBase(this, _serviceName)[_serviceName]] = reducer, _extends2));
    savedReducers = allReducers;
    this.getStore().replaceReducer(combineReducers(allReducers));
  };
  _proto.select = function select() {
    return this.getState()[_classPrivateFieldLooseBase(this, _serviceName)[_serviceName]];
  };
  _proto.dispatchAction = function dispatchAction(newStateFn, trace) {
    if (trace === void 0) {
      trace = 'dispatchAction';
    }
    this.dispatch({
      type: _classPrivateFieldLooseBase(this, _serviceName)[_serviceName] + "::" + TRACE_KEY + "::" + trace,
      params: [newStateFn, trace]
    });
  };
  _proto.resetState = function resetState() {
    this.dispatch({
      type: _classPrivateFieldLooseBase(this, _serviceName)[_serviceName] + "::resetState"
    });
  };
  _proto.createSelector = function createSelector$1() {
    return createSelector.apply(void 0, arguments)(this.getState());
  };
  return BaseReducer;
}(WithStore);

var _services = /*#__PURE__*/_classPrivateFieldLooseKey("services");
var _container = /*#__PURE__*/_classPrivateFieldLooseKey("container");
var ReduxDIContainer = /*#__PURE__*/function () {
  function ReduxDIContainer() {
    Object.defineProperty(this, _services, {
      writable: true,
      value: {}
    });
    Object.defineProperty(this, _container, {
      writable: true,
      value: createContainer({
        injectionMode: InjectionMode.PROXY
      })
    });
  }
  var _proto = ReduxDIContainer.prototype;
  _proto.registerServices = function registerServices(services) {
    var registrations = services.reduce(function (acc, curr) {
      var _extends2;
      return _extends({}, acc, (_extends2 = {}, _extends2[curr.key] = asClass(curr["class"], {
        lifetime: Lifetime.SINGLETON
      }), _extends2));
    }, {});
    _classPrivateFieldLooseBase(this, _container)[_container].register(registrations);
  };
  _proto.injectStore = function injectStore(store, initialReducers) {
    var _this = this;
    var registrations = _classPrivateFieldLooseBase(this, _container)[_container].registrations;
    Object.keys(registrations).forEach(function (key) {
      var service = _classPrivateFieldLooseBase(_this, _container)[_container].resolve(key);
      if (service instanceof WithStore) {
        service.setStore(store);
      }
      if (service instanceof BaseReducer) {
        service.replaceReducer(key, initialReducers);
      }
      _classPrivateFieldLooseBase(_this, _services)[_services][key] = service;
    });
  };
  _proto.getService = function getService(name) {
    return _classPrivateFieldLooseBase(this, _container)[_container].resolve(name);
  };
  return ReduxDIContainer;
}();

export { BaseReducer, ReduxDIContainer, WithStore };
//# sourceMappingURL=index.modern.js.map
